package main

import (
	"context"
	"encoding/hex"
	"errors"
	"flag"
	"fmt"
	"github.com/Shopify/sarama"
	"github.com/golang/protobuf/proto"
	"strconv"
	"strings"
	"sub_demo/proto/subscribe"
	"sub_demo/sasl_sarama"
	"time"
)

var (
	partitionConsumers = make(map[int32]*partitionMsgConsumer)
	brokerList         = flag.String("brokers", "", "REQUIRED: The comma separated list of brokers in the Kafka cluster")
	topic              = flag.String("topic", "", "REQUIRED: the topic to consume")
	group              = flag.String("group", "", "REQUIRED: the consumer group")
	user               = flag.String("user", "", "the user name")
	password           = flag.String("password", "", "the password")
	trans2sql          = flag.Bool("trans2sql", false, "trans message to sql or not")
)

// 本Demo使用sarama框架消费Kafka数据，基本流程为：
//1. 创建 Kafka 消费者和消费组，在一个无限循环中调用消费组的 Consume 方法不停消费消息；
//2. 依次遍历消息，并根据消息中的分区找到分区对应的 partitionMsgConsumer 对象，由该对象对消息进行处理；
//3. partitionMsgConsumer 检查消息头的序列号，如果发现重复消息则丢弃，如果发现消息丢失则终止运行；
//4. partitionMsgConsumer 将原始消息反序列化为 Envelope 结构，如果 Envelope.Data 存在分段则进行拼接；
//5. 将 Envelope.Data 用 Protobuf 解码为 Entries；
//6. 遍历 Entries 中的每一个 Entry，将数据打印出来。
func main() {
	flag.Parse()
	config := sarama.NewConfig()
	config.Version = sarama.V0_11_0_0
	config.Consumer.Return.Errors = true
	config.Consumer.Offsets.Initial = sarama.OffsetOldest
	if *user != "" {
		config.Net.SASL.Enable = true
		config.Net.SASL.User = *user
		config.Net.SASL.Password = *password
		config.Net.SASL.Handshake = true
		config.Net.SASL.SCRAMClientGeneratorFunc = func() sarama.SCRAMClient { return &sasl_sarama.XDGSCRAMClient{HashGeneratorFcn: sasl_sarama.SHA512} }
		config.Net.SASL.Mechanism = sarama.SASLTypeSCRAMSHA512
	}

	client, err := sarama.NewClient(strings.Split(*brokerList, ","), config)
	if err != nil {
		panic(err)
	}
	defer func() { _ = client.Close() }()

	group, err := sarama.NewConsumerGroupFromClient(*group, client)
	if err != nil {
		panic(err)
	}
	defer func() { _ = group.Close() }()

	// Track errors
	go func() {
		for err := range group.Errors() {
			fmt.Println("ERROR", err)
		}
	}()

	// Iterate over consumer sessions.
	ctx := context.Background()
	for {
		handler := demoConsumerGroupHandler{}
		// Consume 方法回先后回调 handler 的 Setup 方法和 ConsumeClaim 方法
		err := group.Consume(ctx, []string{*topic}, handler)
		if err != nil {
			panic(err)
		}
	}
}

type demoConsumerGroupHandler struct {
}

func (demoConsumerGroupHandler) Setup(sess sarama.ConsumerGroupSession) error {
	setup := make(map[int32]bool)
	for _, part := range sess.Claims()[*topic] {
		if _, ok := partitionConsumers[part]; !ok {
			partitionConsumers[part] = &partitionMsgConsumer{}
		}

		setup[part] = true
	}

	var invalid []int32
	for p := range partitionConsumers {
		if _, ok := setup[p]; !ok {
			invalid = append(invalid, p)
		}
	}

	for _, p := range invalid {
		delete(partitionConsumers, p)
	}

	return nil
}

func (demoConsumerGroupHandler) Cleanup(_ sarama.ConsumerGroupSession) error { return nil }

func (h demoConsumerGroupHandler) ConsumeClaim(sess sarama.ConsumerGroupSession, claim sarama.ConsumerGroupClaim) error {
	for msg := range claim.Messages() {
		if err := partitionConsumers[msg.Partition].consume(sess, msg); nil != err {
			return err
		}
	}
	return nil
}

type partitionMsgConsumer struct {
	completeMsg       subscribe.Envelope
	lastPartitionSeq  uint64
	lastPartitionSeq2 uint64
}

func (pmc *partitionMsgConsumer) consume(sess sarama.ConsumerGroupSession, msg *sarama.ConsumerMessage) error {
	// 每个消息的消息头部都有一个编号记做ps，这个编号在同一个分区内正常情况下保证单调递增且连续
	// ps2是ps的升级版，2022.11.1之后启动的订阅任务包含此属性，可以最大限度地解决Kafka消息重复投递的问题，
	// ps解决的是Kafka分区中消息重复投递(Kafka自身机制导致)的问题，ps2在ps的基础上还可以解决后端任务重启后
	// 重复解析部分binlog导致的数据重复投递问题。
	ps2, err := getPartitionSeq2(msg)
	var validPs uint64
	if err == nil {
		if pmc.lastPartitionSeq2 > 0 {
			if ps2 < pmc.lastPartitionSeq2 {
				fmt.Printf("duplicated msg, partition:%v, offset:%v, maxSeq:%v, current:%v\n",
					msg.Partition, msg.Offset, pmc.lastPartitionSeq2, ps2)
				return nil
			} else if ps2 == pmc.lastPartitionSeq2 { // 可能是重复消息也可能是checkpoint或heartbeat消息
				if !isCheckpoint(msg) && !isHeartbeat(msg) {
					fmt.Printf("duplicated msg, partition:%v, offset:%v, maxSeq:%v, current:%v\n",
						msg.Partition, msg.Offset, pmc.lastPartitionSeq2, ps2)
					return nil
				}
			} else if ps2 != pmc.lastPartitionSeq2+1 {
				// 在极端情况下，kafka也可能出现丢失消息的情况（kafka自身机制的原因，概率极小），此时partitionSeq不连续，
				// 如果对数据一致性要求较高，需要停止消费数据，然后在控制台重置订阅任务再重新配置。
				// 如果对数据一致性要求不高，可以接受个别消息丢失，那么跳过此条件判断即可。
				msg := fmt.Sprintf("the PartitionSeq2 is discontinuous, last: %v, current: %v, "+
					"partition:%v, offset:%v", pmc.lastPartitionSeq2, ps2, msg.Partition, msg.Offset)
				panic(msg)
			}
		}
		pmc.lastPartitionSeq2 = ps2
		validPs = ps2
	} else { // 之前的任务没有ps2这个属性，会走到这里
		ps := getPartitionSeq(msg)
		if ps <= pmc.lastPartitionSeq {
			fmt.Printf("duplicated msg: %+v\n", msg)
			return nil
		} else if ps != pmc.lastPartitionSeq+1 && pmc.lastPartitionSeq != 0 {
			msg := fmt.Sprintf("the PartitionSeq is discontinuous, last: %v, current: %v, "+
				"partition:%v, offset:%v", pmc.lastPartitionSeq, ps, msg.Partition, msg.Offset)
			panic(msg)
		}
		pmc.lastPartitionSeq = ps
		validPs = ps
	}

	// 将kafka消息的Value值转换为Envelope
	envelope := subscribe.Envelope{}
	err = proto.Unmarshal(msg.Value, &envelope)
	if nil != err {
		fmt.Println(err)
		return err
	}
	// 当前Envelope的协议版本为1
	if 1 != envelope.Version {
		fmt.Printf("unsupported version: %v, message: %+v\n", envelope.Version, envelope)
		return errors.New("unsupported version")
	}

	//Envelope的Data字段对应Entries二进制序列，Entries包含一个或多个Event。在封装消费的过程中，
	//当一个Event的大小超过单个kafka消息大小限制时，Entries过大，需要将Entries二进制序列分割为多段，
	//每段装入一个Envelope，total和index字段分别记录总段数和分段序号，所以这里需要将分段的Data做拼接。
	if envelope.Index == 0 {
		pmc.completeMsg = envelope
	} else {
		pmc.completeMsg.Data = append(pmc.completeMsg.Data, envelope.Data...)
	}
	// index starts from zero
	if envelope.Index < envelope.Total-1 {
		return nil
	}

	// 将Envelope.Data反序列化为Entries
	entries := subscribe.Entries{}
	err = proto.Unmarshal(pmc.completeMsg.Data, &entries)
	if nil != err {
		fmt.Println(err)
		return err
	}

	// 遍历Entries中的Entry，在onEntry函数中处理消费到的每个Entry
	for _, entry := range entries.Items {
		if err = onEntry(msg.Partition, msg.Offset, validPs, entry); nil != err {
			fmt.Println(err)
			return err
		}
	}

	// 当消费到CHECKPOINT消息时，做一次提交。CHECKPOIN消息在entries中单独存在，所以取Items[0]判断即可。
	if len(entries.Items) > 0 && entries.Items[0].Header.MessageType == subscribe.MessageType_CHECKPOINT {
		sess.MarkMessage(msg, "")
	}

	return nil
}

const partitionSeqKey = "ps"
const partitionSeq2Key = "ps2"

func getPartitionSeq(msg *sarama.ConsumerMessage) uint64 {
	for _, m := range msg.Headers {
		if string(m.Key) == partitionSeqKey {
			n, err := strconv.ParseUint(string(m.Value), 10, 64)
			if nil != err {
				panic("unexpected msg seq: " + string(m.Value))
			}
			return n
		}
	}
	panic(partitionSeqKey + " does not exists")
}

func getPartitionSeq2(msg *sarama.ConsumerMessage) (uint64, error) {
	for _, m := range msg.Headers {
		if string(m.Key) == partitionSeq2Key {
			n, err := strconv.ParseUint(string(m.Value), 10, 64)
			if nil != err {
				panic("unexpected msg seq2: " + string(m.Value))
			}
			return n, nil
		}
	}
	return 0, errors.New(partitionSeq2Key + " does not exists")
}

// 如果传入trans2sql参数，将Entry里的数据解析为sql语句并打印出来，否则直接将Entry打印出来。
// 您可以在此函数基础上开发满足自己需要的数据处理逻辑。
func onEntry(partitionNum int32, offset int64, partitionSeq uint64, entry *subscribe.Entry) error {
	if *trans2sql {
		return trans2SQL(partitionNum, offset, partitionSeq, entry)
	} else {
		fmt.Printf("partition: %v, offset: %v, partitionSeq: %v, %+v\n", partitionNum, offset, partitionSeq, entry)
		return nil
	}
}

func trans2SQL(partitionNum int32, offset int64, partitionSeq uint64, entry *subscribe.Entry) error {
	header := entry.Header
	fmt.Printf("-->[partition: %d, offset: %d, partitionSeq: %d] [%s:%d], happenedAt: %s\n",
		partitionNum, offset, partitionSeq, header.FileName, header.Position, time.Unix(int64(header.Timestamp), 0))

	switch entry.Header.MessageType {
	case subscribe.MessageType_HEARTBEAT:
		fmt.Println("heartbeat:", entry.Event.HeartbeatEvent.Epoch)
	case subscribe.MessageType_CHECKPOINT:
		evt := entry.Event.CheckpointEvent
		fmt.Println("Checkpoint: gtid=", evt.SyncedGtid, ", file=", evt.FileName, ", position=", evt.Position)
	case subscribe.MessageType_BEGIN:
		fmt.Println("BEGIN")
	case subscribe.MessageType_COMMIT:
		fmt.Println("COMMIT")
	case subscribe.MessageType_ROLLBACK:
		fmt.Println("ROLLBACK")
	case subscribe.MessageType_DDL:
		fmt.Println(entry.Event.DdlEvent.Sql)
	case subscribe.MessageType_DML:
		dmlEvt := entry.Event.DmlEvent
		switch dmlEvt.DmlEventType {
		case subscribe.DMLType_INSERT:
			if err := transInsert(entry); nil != err {
				return err
			}
		case subscribe.DMLType_DELETE:
			if err := transDelete(entry); nil != err {
				return err
			}
		case subscribe.DMLType_UPDATE:
			if err := transUpdate(entry); nil != err {
				return err
			}
		default:
			return errors.New("unsupported dml event type")
		}
	default:
		return errors.New("unsupported message type")
	}

	return nil
}

func transInsert(entry *subscribe.Entry) error {
	dmlEvt := entry.Event.DmlEvent
	fmt.Printf("INSERT INTO `%v`.`%v` VALUES ", entry.Header.SchemaName, entry.Header.TableName)
	rowSep := ""
	for _, row := range dmlEvt.Rows {
		fmt.Print(rowSep)
		fmt.Print("(")
		sep := ""
		for _, col := range row.NewColumns {
			fmt.Print(sep)
			if d, err := decode(col); nil != err {
				fmt.Println(err)
				return err
			} else {
				fmt.Printf("%v", d)
			}
			sep = ", "
		}

		fmt.Print(")")
		rowSep = ", "
	}

	fmt.Println("")
	return nil
}

func transDelete(entry *subscribe.Entry) error {
	dmlEvt := entry.Event.DmlEvent
	header := fmt.Sprintf("DELETE FROM `%v`.`%v` WHERE ", entry.Header.SchemaName, entry.Header.TableName)
	for _, row := range dmlEvt.Rows {
		fmt.Print(header)
		sep := ""
		for idx, col := range row.OldColumns {
			if col.DataType == subscribe.DataType_NA {
				continue
			}

			fmt.Print(sep)
			colDef := dmlEvt.Columns[idx]
			fmt.Printf("`%v`", colDef.Name)
			if col.DataType == subscribe.DataType_NIL {
				fmt.Print(" IS NULL")
			} else if d, err := decode(col); nil != err {
				fmt.Println(err)
				return err
			} else {
				fmt.Printf(" = %v", d)
			}
			sep = " AND "
		}

		fmt.Println(" LIMIT 1")
	}

	return nil
}

func transUpdate(entry *subscribe.Entry) error {
	dmlEvt := entry.Event.DmlEvent
	header := fmt.Sprintf("UPDATE `%v`.`%v` SET ", entry.Header.SchemaName, entry.Header.TableName)
	for _, row := range dmlEvt.Rows {
		fmt.Print(header)
		sep := ""
		for idx, col := range row.NewColumns {
			if col.DataType == subscribe.DataType_NA {
				continue
			}

			fmt.Print(sep)
			colDef := dmlEvt.Columns[idx]
			fmt.Printf("`%v`", colDef.Name)
			if d, err := decode(col); nil != err {
				fmt.Println(err)
				return err
			} else {
				fmt.Printf(" = %v", d)
			}
			sep = ", "
		}

		fmt.Print(" WHERE ")
		sep = ""
		for idx, col := range row.OldColumns {
			if col.DataType == subscribe.DataType_NA {
				continue
			}

			fmt.Print(sep)
			colDef := dmlEvt.Columns[idx]
			fmt.Printf("`%v`", colDef.Name)
			if col.DataType == subscribe.DataType_NIL {
				fmt.Print(" IS NULL")
			} else if d, err := decode(col); nil != err {
				fmt.Println(err)
				return err
			} else {
				fmt.Printf(" = %v", d)
			}
			sep = " AND "
		}

		fmt.Println(" LIMIT 1")
	}

	return nil
}

func decode(data *subscribe.Data) (string, error) {
	switch data.DataType {
	case subscribe.DataType_INT8, subscribe.DataType_UINT8,
		subscribe.DataType_INT16, subscribe.DataType_UINT16,
		subscribe.DataType_INT32, subscribe.DataType_UINT32,
		subscribe.DataType_INT64, subscribe.DataType_UINT64,
		subscribe.DataType_FLOAT32, subscribe.DataType_FLOAT64,
		subscribe.DataType_DECIMAL:
		return data.Sv, nil
	case subscribe.DataType_STRING:
		return "_binary'" + string(data.Bv) + "'", nil
	case subscribe.DataType_BYTES:
		return "x'" + hex.EncodeToString(data.Bv) + "'", nil
	case subscribe.DataType_NA:
		return "DEFAULT", nil
	case subscribe.DataType_NIL:
		return "NULL", nil
	default:
		return "", errors.New("unsupported data type")
	}
}

func isCheckpoint(msg *sarama.ConsumerMessage) bool {
	for _, m := range msg.Headers {
		if string(m.Key) == "Checkpoint" {
			return true
		}
	}
	return false
}

func isHeartbeat(msg *sarama.ConsumerMessage) bool {
	for _, m := range msg.Headers {
		if string(m.Key) == "Heartbeat" {
			return true
		}
	}
	return false
}
